/* Copyright 2015, 2016, 2017 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Some alternate paperfolding curve graph stuff using Nauty. */

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* GP-DEFINE  read("alternate-defines.gp"); */


/* ------------------------------------------------------------------------- */
/* Misc Helpers */

/* bit position of the highest 1 bit of an int n>=0 */
static const int  int_highbit_pos = sizeof(int)*CHAR_BIT - 2;

/* Return a bit mask "000011111" which is the lowest 0-bit or 1-bit
   (respectively) of n and all the bits below that position. */
#define ALTERNATE_MASK_LOWEST_ZERO_AND_BELOW(n)   ((n) ^ ((n) + 1))
#define ALTERNATE_MASK_LOWEST_ONE_AND_BELOW(n)    ((n) ^ ((n) - 1))

/* Return a bit mask "000100000" which is the bit above the lowest 0-bit of
   n. */
#define ALTERNATE_MASK_BIT_ABOVE_LOWEST_ZERO(n)    \
  (ALTERNATE_MASK_LOWEST_ZERO_AND_BELOW(n) + 1)

/* Return a bit mask "000100000" which is the bit above the lowest 1-bit of
   n. */
#define ALTERNATE_MASK_BIT_ABOVE_LOWEST_ONE(n)    \
  (ALTERNATE_MASK_LOWEST_ONE_AND_BELOW(n) + 1)

/* Change (x,y) to (-y,x) which is a rotation by +90 degrees. */
#define ALTERNATE_XY_ROTATE_PLUS90(x,y)            \
  do {                                          \
    int ALTERNATE_XY_ROTATE_PLUS90__tmp = (x);     \
    (x) = -(y);                                 \
    (y) = ALTERNATE_XY_ROTATE_PLUS90__tmp;         \
  } while (0)

/* Change (x,y) to (y,-x) which is a rotation by -90 degrees. */
#define ALTERNATE_XY_ROTATE_MINUS90(x,y)           \
  do {                                          \
    int ALTERNATE_XY_ROTATE_MINUS90__tmp = (x);    \
    (x) = (y);                                  \
    (y) = - ALTERNATE_XY_ROTATE_MINUS90__tmp;      \
  } while (0)

/* ------------------------------------------------------------------------- */
/* Alternate Sequence Values */

void
alternate_overflow (void)
{
  fprintf(stderr, "alternate recurrence overflow\n");
  abort();
}

/* alternate_P(k) returns the number distinct points visited by alternate
   curve k.
*/
int
alternate_P (int k)
{
  /* 1/2*(2^k + vector_modulo([3,4],k)*2^floor(k/2)) */
  return (k <= 1 ? k+2
          : (1<<(k-1)) + ((k&1?4:3) << ((k>>1)-1)));
}

/* alternate_TP(k) returns the number of distinct points visited by twin
   alternate k.
*/
int
alternate_TP (int k)
{
  /* 2*2^k + vector_modulo([3,4],k) * 2^floor(k/2)  - 1) */
  return (1<<(k+1)) + ((k&1?4:3) << (k>>1)) - 1;
}

/* ------------------------------------------------------------------------- */
/* Twin Alternate Area Tree Sequences */

/* alternate_TAdiameter(k) returns the diameter of twin_alternate area tree k.
 */
int
alternate_TAdiameter(int k)
{
  /* vector_modulo([7,10],k) * 2^floor(k/2) - 2*k - 7 */
  return ((k&1 ? 10 : 7) << (k>>1) ) - 2*k  - 7;
}

/* alternate_TAW(k) returns the Wiener index of twin_alternate area
   tree k.

   vector_modulo([15/14, 43/28],k) * 4^k * 2^floor(k/2)
   - (1/4*k + 1) * 4^k
   - 1/14 * 2^k
*/
unsigned long long
alternate_TAW(int k)
{
  return (((k&1 ? 43ULL : 30ULL) << ((5*k)>>1))
          - (1ULL<<(k+1)) ) / 28
    - (k==0 ? 1 : ((k+4ULL) << ((k<<1)-2)));
}

/* Return the vertex above n in the twin alternate area tree, or -1 if no
   edge in that direction.
 */
int
alternate_TAVertexToUpper (int n)
{
  int mask = ALTERNATE_MASK_LOWEST_ZERO_AND_BELOW(n);
  if(n & (mask+1)) return -1;
  else             return n + mask + 2;
}
/* Return the vertex below n in the twin alternate area tree, or -1 if no
   edge in that direction.
 */
int
alternate_TAVertexToLower (int n)
{
  int mask = ALTERNATE_MASK_LOWEST_ONE_AND_BELOW(n);
  if(n & (mask+1)) return n - mask - 2;
  else             return -1;
}
/* Return the vertex right of n in the twin alternate area tree, or -1 if no
   edge in that direction
 */
int
alternate_TAVertexToRight (int n)
{
  if (n&1) return n ^ 1;
  else     return -1;
}
/* Return the vertex left of n in the twin alternate area tree, or -1 if no
   edge in that direction.
 */
int
alternate_TAVertexToLeft (int n)
{
  if (n&1) return -1;
  else     return n ^ 1;
}

/* Return the vertex in direction dir mod 4 from n in the twin alternate
   area tree, or -1 if no edge in that direction.
 */
int
alternate_TAstepDir (int n, int dir)
{
  switch (dir & 3) {
  case 0:  return alternate_TAVertexToRight(n);  /* dir=0 */
  case 1:  return alternate_TAVertexToUpper(n);  /* dir=1 */
  case 2:  return alternate_TAVertexToLeft(n);   /* dir=2 */
  default: return alternate_TAVertexToLower(n);  /* dir=3 */
  }
}

/* new_dir = alternate_TAparentDir_state_table[dir][bit]
   for bits of n taken high to low
*/
const int alternate_TAparentDir_state_table[4][2]
  = { { 2, 3 },     /* 0=right */
      { 2, 1 },     /* 1=upper */
      { 1, 0 },     /* 2=left */
      { 3, 0 } };   /* 3=lower */

const int alternate_TAparentDir_starting_dir = 3;

/* Return direction 0,1,2,3 which is the parent of vertex n>=1 of the
   twin alternate area tree.  n=0 is the root (so has no parent).
*/
int
alternate_TAparentDir (int n)
{
  assert (n >= 0);
  if (n==0) return -1;
  int dir = alternate_TAparentDir_starting_dir;
  /* printf("alternate_TAparentDir() n=%d\n", n); */
  int pos;
  for (pos = int_highbit_pos; pos >= 0; pos--) {
    int bit = (n >> pos) & 1;
    /* printf(" dir=%d pos=%d bit %d\n", dir, pos, bit); */
    dir = alternate_TAparentDir_state_table[dir][bit];
  }
  /* printf(" final dir=%d\n", dir); */
  return dir;
}

/* Return the vertex which is the parent of vertex n>=1 of the twin alternate
   area tree.  n=0 is the root (so has no parent).
*/
int
alternate_TAparent (int n)
{
  if (n == 0) { return -1; }
  int dir = alternate_TAparentDir(n);
  assert (dir >= 0);
  n = alternate_TAstepDir(n, dir);
  assert (n >= 0);
  return n;
}

/* ------------------------------------------------------------------------- */
/* Alternate Curve Graph */

/* alternate_make_graph(g, m,n, k) sets graph g to alternate curve k.
   n should be the number of vertices alternate_P(k).
   g should be already allocated to size m,n.
*/
void
alternate_make_graph(graph *g, int m, int n, int k)
{
  DEBUG(printf("alternate_make_graph() n=%d  k=%d\n", n,k));
  int xhalf = (1 << ( k   >>1));
  int xmax  = (1 << ((k+1)>>1)) - (k&1);
  DEBUG(printf(" xmax %d  xhalf %d\n", xmax, xhalf));

  assert(n == alternate_P(k));
  assert(m >= SETWORDSNEEDED(n));
  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g,m,n);

  int v = 0;
  int x;
  for (x = 0; x <= xmax; x++) {
    int ymax = (x <= xhalf ? x : xmax - x + 1);
    if (x == xmax && ! (k&1)) ymax--;
    DEBUG(printf(" x=%d ymax %d\n", x, ymax));

    int next_ymax = ((x+1) <= xhalf ? (x+1) : xmax - (x+1) + 1);

    int y;
    for (y = 0; y <= ymax; y++) {

      /* horizontal right */
      if (x < xmax && ! (y==0 && (x&1)) && y <= next_ymax) {
        int to = v + ymax + 1;
        DEBUG(printf("  y=%d right %d -- %d\n", y, v, to));
        assert (to < n);
        assert (! ISELEMENT(GRAPHROW(g,v,m),to));
        assert (! ISELEMENT(GRAPHROW(g,to,m),v));
        ADDONEEDGE(g, v,to, m);
      }

      /* vertical up */
      if (y < ymax && ! (x==xmax && (y&1))) {
        int to = v + 1;
        DEBUG(printf("  y=%d up %d -- %d\n", y, v, to));
        assert (to < n);
        assert (! ISELEMENT(GRAPHROW(g,v,m),to));
        assert (! ISELEMENT(GRAPHROW(g,to,m),v));
        ADDONEEDGE(g, v,to, m);
      }

      v++;
    }
  }
  DEBUG(printf(" end %d (cf n=%d)\n", v, n));
  assert (v == n);
}

/* ------------------------------------------------------------------------- */
/* Twin Alternate Graph */

/* Set graph g to twin_alternate k.
   n should be the number of vertices alternate_TP(k).
   g should be already allocated to size m,n.
*/
void
alternate_make_twin_alternate_graph(graph *g, int m, int n, int k)
{
  DEBUG(printf("alternate_make_twin_alternate_graph() n=%d  k=%d\n", n,k));
  int xmax = (1 << ((k+2) >>1));
  int ymax = (1 << ((k+1) >>1));
  DEBUG(printf(" xmax %d  ymax %d\n", xmax, ymax));

  assert(n == alternate_TP(k));
  assert(m >= SETWORDSNEEDED(n));
  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g,m,n);

  int v = 0;
  int y;
  for (y = 0; y <= ymax; y++) {
    int x_lo = -y + (y == ymax);
    int x_hi = -y + xmax - (y == 0);
    DEBUG(printf(" y=%d x range %d to %d  at v=%d\n", y, x_lo, x_hi, v));

    int x;
    for (x = x_lo; x <= x_hi; x++) {

      /* horizontal right */
      if (x < x_hi
          && ! (y==0 && (x&1))
          && ! (y==ymax && !(x&1))) {
        int to = v + 1;
        DEBUG(printf("  x=%d right %d -- %d\n", x, v, to));
        assert (to < n);
        assert (! ISELEMENT(GRAPHROW(g,v,m),to));
        assert (! ISELEMENT(GRAPHROW(g,to,m),v));
        ADDONEEDGE(g, v,to, m);
      }

      /* vertical up */
      if (y < ymax
          && x < (y==0 ? xmax : x_hi)
          && ! (x==xmax && (y&1))) {
        int to = v + xmax;
        DEBUG(printf("  x=%d up %d -- %d\n", x, v, to));
        assert (to < n);
        assert (! ISELEMENT(GRAPHROW(g,v,m),to));
        assert (! ISELEMENT(GRAPHROW(g,to,m),v));
        ADDONEEDGE(g, v,to, m);
      }

      v++;
    }
  }
  DEBUG(printf(" end %d (cf n=%d)\n", v, n));
  assert (v == n);
}

/* ------------------------------------------------------------------------- */
/* Twin Alternate Area Tree */

/* twin_alternate_area_tree_n(k) returns the number of vertices in
   twin_alternate area tree k.  This is simply 2^k.

   twin_alternate_area_tree(g, m,n) sets graph g to twin_alternate area tree k.
   n should be the number of vertices as per twin_alternate_area_tree_n(k).
   g should be already allocated to size m,n.

   The twin_alternate area tree is quite sparse, since each vertex degree <= 3,
   so usually twin_alternate_area_tree_sg() below is better, when sparsegraph
   suits.
*/
int
twin_alternate_area_tree_n(int k)
{
  int n = 1 << k;
  if (n <= 0) { alternate_overflow(); }
  return n;
}
void
alternate_make_TA(graph *g, int m, int n)
{
  int from;

  assert (n >= 0);
  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  for (from = 0; from < n; from++) {

    /* horizontal from=xx1 edge to=xx0 */
    if (from & 1) {
      int to = from ^ 1;
      assert (! ISELEMENT(GRAPHROW(g,from,m),to));
      assert (! ISELEMENT(GRAPHROW(g,to,m),from));
      ADDONEEDGE(g, from,to, m);
    }

    /* to lower */
    if (from <= n - 1) {    /* no zeros in from=1111 */
      int mask = ALTERNATE_MASK_LOWEST_ONE_AND_BELOW(from);
      if (from & (mask+1)) {       /* from=...11000 */
        int to = from - mask - 2;  /*   to=...00111 */
        assert (! ISELEMENT(GRAPHROW(g,from,m),to));
        assert (! ISELEMENT(GRAPHROW(g,to,m),from));
        ADDONEEDGE(g, from,to, m);
      }
    }
  }
}

/* alternate_make_TA_sg(g,k) sets sparsegraph g to twin_alternate
   area tree level k.  Any previous contents of g are cleared and
   dynamically reallocated if necessary for k.  (A new sg should be SG_DECL()
   in the usual way.)

   Edges are added in both directions between vertices, as usual for an
   undirected sparsegraph.
*/
void
alternate_make_TA_sg(sparsegraph *sg, int k)
{
  DEBUG(printf("alternate_make_TA_sg()  k=%d\n", k));
  assert (k >= 0);
  int num_vertices = twin_alternate_area_tree_n(k);
  int num_edges = num_vertices - 1;

  int nde = 2*num_edges;
  SG_ALLOC(*sg, num_vertices, nde, "malloc");
  sg->nv = num_vertices;
  sg->nde = nde;

  int *dd = sg->d;
  size_t *vv = sg->v;
  int *ee = sg->e;
  int epos = 0;
  int from;
  for (from = 0; from < num_vertices; from++) {
    *vv++ = epos;
    int from_degree = 0;

    /* horizontal from=...0 edge to=...1 or the other way around */
    if (k) {
      int to = from ^ 1;
      DEBUG(printf("  horizontal 0x%X to 0x%X\n", from, to));
      assert(epos<nde);
      ee[epos++] = to;
      from_degree++;
    }

    /* to lower */
    if (from > 0) {   /* no 1s in from=00000 */
      int mask = ALTERNATE_MASK_LOWEST_ONE_AND_BELOW(from);
      if (from & (mask+1)) {    /* from=...11000 */
        int to = from - mask - 2;   /*   to=...00111 */
        DEBUG(printf("  to lower 0x%X to 0x%X\n", from, to));
        assert(epos<nde);
        ee[epos++] = to;
        from_degree++;
      }
    }

    /* to upper */
    int mask = ALTERNATE_MASK_LOWEST_ZERO_AND_BELOW(from);
    if (from < num_vertices - 1) {
      if ((from & (mask+1)) == 0) {   /* from=...00111 */
        int to = from + mask + 2;     /*   to=...11000 */
        if (to < num_vertices) {
          DEBUG(printf("  to upper 0x%X to 0x%X\n", from, to));
          assert(epos<nde);
          ee[epos++] = to;
          from_degree++;
        }
      }
    }

    *dd++ = from_degree;
  }
  DEBUG(printf("  final epos=%d cf nde=%d\n",epos,nde));
  assert (epos == nde);
}

/* Set vpar[1..2^k] to twin alternate area tree level k.
   The vertex numbers are the area tree 0 .. 2^k-1 and each +1.
*/
void
alternate_make_TA_vpar(int *vpar, int k)
{
  assert (k >= 0);
  int n_max = twin_alternate_area_tree_n(k);
  int n;
  vpar[0] = 0;
  for (n = 1; n <= n_max; n++)
    vpar[n] = alternate_TAparent(n-1) + 1;
}

/* ------------------------------------------------------------------------- */

void
test (void)
{
  int max_bad = 10;
  int bad = 0;
  printf ("int_highbit_pos = %d\n", int_highbit_pos);
  {
    /* alternate_P */
    int want[] = { 2, 3, 5, 8, 14, 24, 44, 80, 152, 288 };
    int k;
    for (k = 0; k < sizeof(want)/sizeof(want[0]); k++) {
      int got = alternate_P(k);
      if (! (got == want[k])) {
        printf ("alternate_P(%d) want %d got %d\n", k, want[k], got);
        if (++bad >= max_bad) abort();
      }
    }
  }
  {
    /* alternate_TAparentDir */
    int want[] = { 0, 2, 3, 1, 0, 3, 0, 2, 1, 2, 3, 3, 0, 2, 3, 1, 0 };
    int i;
    for (i = 0; i < sizeof(want)/sizeof(want[0]); i++) {
      int n = i+1;
      int got = alternate_TAparentDir(n);
      if (! (got == want[i])) {
        printf ("alternate_TAparentDir(%d) want %d got %d\n", n, want[i], got);
        if (++bad >= max_bad) abort();
      }
    }
  }
  {
    /* alternate_TAparent */
    int want[] = { 0, 3, 0, 7, 4, 1, 6, 9, 14, 11, 8, 3, 12, 15, 12, 19, 16 };
    int i;
    for (i = 0; i < sizeof(want)/sizeof(want[0]); i++) {
      int n = i+1;
      int got = alternate_TAparent(n);
      if (! (got == want[i])) {
        printf ("alternate_TAparent(%d) want %d got %d\n", n, want[i], got);
        if (++bad >= max_bad) abort();
      }
    }
  }
  {
    /* alternate_make_TA() vs alternate_make_TA_sg() vs alternate_make_TA_vpar()
     */
    const int k_max = 8;
    int vpar[(1 << k_max)+1];
    DYNALLSTAT(graph,g,g_sz);
    SG_DECL(sg);
    SG_INIT(sg);
    int k;
    for (k=0; k < 8; k++) {
      alternate_make_TA_vpar(vpar,k);
      int n = twin_alternate_area_tree_n(k);

      alternate_make_TA_sg(&sg,k);
      if (! nautyextra_is_isomorphic_sg_vpar(&sg,vpar,n)) {
        printf ("alternate_make_TA_sg() vs alternate_make_TA_vpar() k=%d\n", k);
        if (++bad >= max_bad) abort();
      }

      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g,g_sz,m,n,"malloc");
      alternate_make_TA(g,m,n);
      if (! nautyextra_is_isomorphic_gsg(g,&sg,m,n)) {
        printf ("alternate_make_TA() vs alternate_make_TA_sg() k=%d\n", k);
        if (++bad >= max_bad) abort();
      }
    }
    SG_FREE(sg);
  }
  if (bad) abort();
}

void
writeautom (int *p, int n)
{
  int i;
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}

void
show_edges (graph *g, int m,int n)
{
  set *row = (set*) g;
  int from, to;
  printf("edges");
  for (from=0; from < n-1; from++, row+=m) {
    for (to=from+1; to < n; to++) {
      if (ISELEMENT(row,to))
        printf(" %d-%d", from,to);
    }
  }
  putchar('\n');
}

void
try_alternate_make_TA (void)
{
  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  statsblk stats;

  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLSTAT(int, map,    map_sz);
  SG_DECL(ta);
  SG_DECL(canon_ta);
  SG_DECL(canon_ta2);

  int k = 8;

  options.getcanon = TRUE;
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;

  printf("make tree\n");
  alternate_make_TA_sg(&ta, k);

  printf("\n");
  printf("as generated\n");
  nautyextra_print_sg(&ta);

  int num_vertices = ta.nv;
  DYNALLOC1(int, lab,    lab_sz,    num_vertices, "malloc");
  DYNALLOC1(int, ptn,    ptn_sz,    num_vertices, "malloc");
  DYNALLOC1(int, orbits, orbits_sz, num_vertices, "malloc");
  DYNALLOC1(int, map,    map_sz,    num_vertices, "malloc");

  printf("\nlabel ta\n");
  sparsenauty(&ta,lab,ptn,orbits,&options,&stats,&canon_ta);

  printf("canonical\n");
  nautyextra_print_sg(&canon_ta);

  sortlists_sg(&ta);

  {
    int num_edges = ta.nde;
    int m = SETWORDSNEEDED(num_vertices);
    printf("fcanon m=%d\n", m);
    /* int loops = loopcount(&ta, m, num_vertices); */
    /* printf("loops %d\n", loops); */
    SG_ALLOC(canon_ta2, num_vertices, num_edges, "canon_ta2");
    fcanonise_inv_sg(&ta, m, num_vertices, &canon_ta2,
                     NULL,  /* fmt colours */
                     NULL,  /* distances_sg,  */ /* invarproc */
                     1,     /* mininvarlevel */
                     1,     /* maxinvarlevel */
                     3,     /* invararg */
                     0);    /* digraph */
    sortlists_sg(&canon_ta2);
  }

  putgraph_sg(stdout, &ta, 79);  /* line length */

  printf("\n");
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&ta);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&canon_ta);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&canon_ta2);

  printf("\n");
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&ta);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&canon_ta);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&canon_ta2);

  printf("aresame  %d %d",
         aresame_sg(&ta,&canon_ta),
         aresame_sg(&canon_ta,&canon_ta2));


  printf ("\nGenerators\n");
  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  printf ("num orbits    %d\n", stats.numorbits);
  printf ("errstatus     %d\n", stats.errstatus);
  printf ("numnodes      %lu\n", stats.numnodes);
  printf ("numbadleaves  %lu\n", stats.numbadleaves);
  printf ("maxlevel      %d\n", stats.maxlevel);

  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);
      allgroup(group,writeautom);
    }
  }
}

void
try_alternate_TAdiameter (void)
{
  DYNALLSTAT(graph,g,g_sz);
  int k;

  for (k=0; k <= 10; k++) {
    int formula = alternate_TAdiameter(k);
    int n = twin_alternate_area_tree_n(k);
    int m = SETWORDSNEEDED(n);

    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    alternate_make_TA(g,m,n);

    int radius, diameter;
    diamstats(g,m,n, &radius, &diameter);

    printf("k=%d  radius %d  diameter %d  %d\n",
           k, radius, diameter, formula);
    /* nautyextra_view (g,m,n); */
  }
}

void
try_alternate_TAW (void)
{
  DYNALLSTAT(graph,g,g_sz);
  int k;

  for (k=0; k < 10; k++) {
    int n = twin_alternate_area_tree_n(k);
    int m = SETWORDSNEEDED(n);
    unsigned long long TAW_calc;
    unsigned long long TAW_formula = alternate_TAW(k);

    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    alternate_make_TA(g,m,n);
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(g,m,n, 0, &info);
    TAW_calc = info.W;

    printf("k=%d  Wiener index %Ld %Ld\n", k, TAW_calc, TAW_formula);
  }
  for (k=0; k <= 15; k++)
    printf("%Ld,", alternate_TAW(k));
}

void
try_sequences (void)
{
  int k;
  /* printf("turn right n=1: "); */
  /* for (n=1; n < 10; n++) { */
  /*   printf(" %d,", alternate_turn_is_right(n)); */
  /* } */
  /* printf("\n"); */

  for (k=0; k < 10; k++) {
    int P = alternate_P(k);
    int TP = alternate_TP(k);
    printf("k=%d  P=%d TP=%d\n", k, P, TP);
  }
}

/* TA indnum 2^(k-1)
*/
void
try_TAindnum (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  int k;
  for (k=0; k < 12; k++) {
    int n = twin_alternate_area_tree_n(k);
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    alternate_make_TA_vpar(vpar, k);
    printf("k=%d n=%d\n", k,n);

    int indnum = nautyextra_indnum_vpar(vpar,n);

    struct nautyextra_indnum_and_count  info;
    nautyextra_indnum_and_count_vpar (vpar,n, &info);
    assert(indnum == info.indnum);

    unsigned long long indsets_count =0;
    /* indsets_count = nautyextra_indsets_count_vpar(vpar,n); */

    printf("    indnum %d count %Lu   indsets %Lu\n",
           indnum, info.count, indsets_count);
  }
}

/* TA domnum 3*2^(k-3) starting 1,1,2
   domnum 1,1,2,3,6,12,24,48,96,192,384,768,1536,3072,6144,12288,
   A098011
*/
void
try_TAdomnum (void)
{
  {
    SG_DECL(sg);
    SG_INIT(sg);
    int k;
    for (k=0; k < 16; k++) {
      struct nautyextra_domnum_and_count  info;
      alternate_make_TA_sg(&sg,k);
      nautyextra_domnum_and_count_sg(&sg,&info);

      printf("k=%d  domnum %d count %Lu\n", k, info.domnum, info.count);
    }
    printf("\n");
    SG_FREE(sg);
  }
  {
    DYNALLSTAT(int,vpar,vpar_sz);
    printf("domnum ");
    int k;
    for (k=0; k < 16; k++) {
      int n = twin_alternate_area_tree_n(k);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      alternate_make_TA_vpar(vpar, k);
      int domnum = nautyextra_domnum_vpar(vpar,n);

      printf("%d,", domnum);
    }
    printf("\n");
  }
}

void
try_twin_alternate_domnum_components (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  int k;
  printf("domnum components ");
  for (k=0; k < 16; k++) {
    int n = twin_alternate_area_tree_n(k);
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    alternate_make_TA_vpar(vpar, k);

    DYNALLSTAT(int,degrees,degrees_sz);
    DYNALLOC1(int,degrees,degrees_sz,n,"nautyextra_check_sg");
    nautyextra_degrees_vpar (vpar,n, degrees);

    int m = SETWORDSNEEDED(n+1);
    DYNALLSTAT(set,absent,absent_sz);
    DYNALLOC1(set,absent,absent_sz,m,"nautyextra_check_sg");
    EMPTYSET(absent,m);
    ADDELEMENT(absent,0);

    int v;
    for (v = 1; v <= n; v++) {
      if (degrees[v] == 1) {
        int p = vpar[v];
        ADDELEMENT(absent,v);
        ADDELEMENT(absent,p);
      }
    }
    int count = 0;
    for (v = 1; v <= n; v++) {
      int p = vpar[v];
      if (! ISELEMENT(absent,v) && ISELEMENT(absent,p)) {
        count++;
      }
    }
    printf("%d,", count);
  }
  printf("\n");
}

void
try_alternate_diameter (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;
  static unsigned long long values[20];

  int k_limit = 11;
  for (k = 0; k <= k_limit; k++) {
    int n = alternate_P(k);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");

    alternate_make_graph(g, m,n, k);
    int diameter_start = 0;
    int diameter_end = (k <= 1 ? n-1
                        : k == 2 ? n-2
                        : (k&1)==0 ? n-1  /* k even top last column */
                        : n-2);           /* k odd bottom last column (of 2) */

    /* nautyextra_view_synchronous = 1; */
    /* nautyextra_view (g,m,n); */

    /* int n = alternate_TP(k); */
    /* int m = SETWORDSNEEDED(n); */
    /* DYNALLOC2(graph,g,g_sz,m,n,"malloc"); */
    /* alternate_make_TA_graph(g, m,n, k); */

    int radius, diameter;
    diamstats(g, m,n, &radius, &diameter);
    int cycles = 0;
    if (n <= WORDSIZE) cycles = cyclecount(g, m,n);

    struct nautyextra_Wiener_info  info;
    nautyextra_Wiener_index(g, m,n, 0, &info);

    int diameter_formula = 0;

    DYNALLSTAT(int,dist,dist_sz);
    DYNALLOC1(int,dist,dist_sz,n,"malloc");
    find_dist((graph*)g, m,n, diameter_start, dist);

    DYNALLSTAT(unsigned long long,count,count_sz);
    DYNALLOC1(unsigned long long,count,count_sz,n,"malloc");
    nautyextra_find_dist_counts((graph*)g, m,n, diameter_start, dist, count);

    printf("k=%d n=%d  radius %d  diameter %d %d  cycles %d\n",
           k, n, radius, diameter, diameter_formula, cycles);
    printf("  W=%Lu (diam %d, %d to %d)\n",
           info.W, info.diameter, info.diameter_from, info.diameter_to);
    printf("  diameter %d to %d dist %d count %Lu\n",
           diameter_start, diameter_end,
           dist[diameter_end], count[diameter_end]);

    if (k<=4) {
      printf("  counts ");
      int v;
      for (v = 0; v < n; v++)
        printf("%Lu,", count[v]);
      printf("\n");
    }
    assert (diameter == dist[diameter_end]);
    values[k] = count[diameter_end];
  }

  for (k = 0; k<=k_limit; k++)
    printf("%Lu,", values[k]);
  printf("\n");
}

void
try_alternate_domnum (void)
{
  DYNALLSTAT(graph, g, g_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;

  for (k = 0; k <= 10; k++) {
    int n = alternate_P(k);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    alternate_make_graph(g, m,n, k);
    nauty_to_sg(g,&sg,m,n);

    struct nautyextra_domnum_and_count  info;
    nautyextra_domnum_and_count_sg (&sg, &info);
    printf("k=%d n=%d  domnum %d count %Lu\n", k, n, info.domnum, info.count);
  }
}

void
try_twin_alternate_domnum (void)
{
  DYNALLSTAT(graph, g, g_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;

  for (k = 0; k <= 8; k++) {
    int n = alternate_TP(k);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    alternate_make_twin_alternate_graph(g, m,n, k);
    nauty_to_sg(g,&sg,m,n);

    struct nautyextra_domnum_and_count  info;
    nautyextra_domnum_and_count_sg (&sg, &info);
    printf("k=%d n=%d  domnum %d count %Lu\n", k, n, info.domnum, info.count);

    unsigned long edges;
    int mindeg;
    int mincount;
    int maxdeg;
    int maxcount;
    boolean eulerian;
    degstats((graph*)g,m,n, &edges, &mindeg, &mincount, &maxdeg, &maxcount,
             &eulerian);
    printf(" edges %lu\n", edges);
  }
}


/* ------------------------------------------------------------------------- */

int
main (void)
{
  setbuf(stdout,NULL);
  /* test(); */

  try_alternate_diameter();
  return 0;
  try_TAdomnum();
  try_TAindnum();
  try_alternate_domnum();
  try_twin_alternate_domnum();
  try_alternate_TAW();
  try_alternate_TAdiameter();
  try_twin_alternate_domnum_components();
  try_sequences();
  return 0;
}
