#!/usr/bin/perl -w

# Copyright 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use List::Util 'max','min','sum';
use Math::BaseCnv 'cnv';

# use Math::PlanePath;
# *_divrem_mutate = \&Math::PlanePath::_divrem_mutate;
# unshift @INC, '/usr/share/perl5';

use FindBin;
use lib "$FindBin::Bin/../tools";
use MyAlternate ':all';
use MyPrintwrap ':all';
use MyFLAT;
$|=1;

# uncomment this to run the ### lines
use Smart::Comments;

sub to_negabinary {
  my ($n) = @_;
  return ($n + 0xAAAAAAAA) ^ 0xAAAAAAAA;
}
sub from_negabinary {
  my ($g) = @_;
  my $p = 1;
  my $n = 0;
  while ($g) {
    if ($g & 1) { $n += $p; }
    $p *= -2;
    $g >>= 1;
  }
  return $n;
}
CHECK {
  foreach my $n (0 .. 256) {
    my $g = to_negabinary($n);
    my $n_again = from_negabinary($g);
    $n == $n_again or die;
  }
}
sub CountNegabinaryOnes {
  my ($n) = @_;
  return CountOneBits(to_negabinary($n));
}
sub negabinary_1s_parity {
  my ($n) = @_;
  return CountNegabinaryOnes($n) & 1;
}

# A069010
sub CountRuns {
  my ($n) = @_;
  my $ret = 0;
  while ($n) {
    $ret += ($n & 3) == 1;
    $n>>=1;
  }
  return $ret;
}
CHECK {
  my @want = (0,1,1,1,1,2,1,1,1,2,2,2,1,2,1,1,1,2,2,2,2,3,2,2,1,
              2,2,2,1,2,1,1,1,2,2,2,2,3,2,2,2,3,3,3,2,3,2,2,1,2,
              2,2,2,3,2,2,1,2,2,2,1,2,1,1,1,2,2,2,2,3,2,2);
  foreach my $n (0 .. $#want) {
    CountRuns($n) == $want[$n] or die;
  }
}
sub CountNegabinaryRuns {
  my ($n) = @_;
  return CountRuns(to_negabinary($n));
}
# not in OEIS: 1,1,1,1,2,2,2,1,2,1,1,1,2,2,2,1,2,2,2,2,3,3,3
# foreach my $n (1 .. 30) {
#   print negabinary_runs_parity($n),",";
# }
# exit;

# A269528
sub negabinary_runs_parity {
  my ($n) = @_;
  return CountNegabinaryRuns($n) & 1;
}
CHECK {
  my @want = (0,1,1,1,1,0,0,0,1,0,1,1,1,0,0,0,1,0,0,0,0,1,1,1,
              0,1,0,0,0,1,0,0,1,0,0,0,0,1,0,0,1,0,1,1,1,0,0,0,
              1,0,0,0,0,1,1,1,0,1,0,0,0,1,0,0,1,0,0,0,0,1,1,1);
  foreach my $n (0 .. $#want) {
    negabinary_runs_parity($n) == $want[$n] or die;
  }
}

my @negabinary_1s_parity_LtoH_table = ([0,0],
                  [3, 4], [4, 3], [1, 6], [2, 5],
                  [4, 7], [3, 8], [6, 5], [5, 6]);
my $negabinary_1s_parity_LtoH_flat = MyFLAT::aref_to_FLAT_DFA
  (\@negabinary_1s_parity_LtoH_table,
   name           => "negabinary 1s parity = 1",
   starting       => 1,
   accepting_list => [2,4,5,8]);


#------------------------------------------------------------------------------

{
  # negabinary 1s parity * 3

  my $f = $negabinary_1s_parity_LtoH_flat
    ->MyFLAT::transmute (initial => 0,
                         func => sub {
                           my ($carry,$bit) = @_;
                           $carry = 3*$bit + $carry;
                           return ($carry>>1, $carry&1);
                         });
  $f = $f->MyFLAT::binary_to_base4 (direction => 'lowtohigh');
  # $f = $f->MyFLAT::blocks (2, direction => 'lowtohigh');
  # $f = $f->MyFLAT::reverse;
  $f = $f->MyFLAT::minimize(verbose=>1);
  $f->MyFLAT::view;
  # MyFLAT::FLAT_print_tikz($f);

  # MyFLAT::FLAT_show_breadth($negabinary_1s_parity_LtoH_flat, 4, 'lowtohigh');
  # MyFLAT::FLAT_show_breadth($f, 4, 'lowtohigh');
  # MyFLAT::FLAT_check_is_equal($f, $negabinary_1s_parity_LtoH_flat);

  my $bad = 0;
  foreach my $n (0 .. 1<<13) {
    my $three_n = 3*$n;
    my $three_n2rev = reverse(cnv($three_n,10,2));
    foreach my $zeros (0 .. 4) {
      my $by_flat = $f->contains($three_n2rev) ? 1 : 0;
      my $by_func = negabinary_1s_parity($n);
      unless ($by_flat == $by_func) {
        print "n=$n 3n=$three_n [$three_n2rev] $by_flat $by_func\n";
        exit if $bad++ > 10;
      }
      $three_n2rev .= '0';
    }
  }

  exit 0;
}

{
  # negabinary 1s parity base 4

  my $f = $negabinary_1s_parity_LtoH_flat
    ->MyFLAT::binary_to_base4 (direction => 'lowtohigh');
  # my $f = $negabinary_1s_parity_LtoH_flat
  #   ->MyFLAT::blocks (2, direction => 'lowtohigh');
  $f = $f->MyFLAT::reverse;
  $f = $f->MyFLAT::minimize(verbose=>1);
  $f->MyFLAT::view;
  MyFLAT::FLAT_print_tikz($f);
  exit 0;
}


#------------------------------------------------------------------------------

{
  # negabinary runs parity
  # carry,pos,prev,runs
  my $cppr_transition = sub {
    my ($cppr, $bit) = @_;
    ### cppr_transition: "$cppr bit $bit"
    my ($carry,$pos,$prev,$runs) = split //, $cppr;
    my $d = $carry + $bit;
    my $out = $d & 1;
    my $new_carry = ($d - $out*(-1)**$pos) / 2;
    $carry == int($carry) or die;
    my $new_prev = $out;
    my $this_run_top = ($out == 0 && $prev == 1);
    my $new_runs = ($this_run_top ? 1-$runs : $runs);
    my $new_pos = 1 - $pos;
    return "$new_carry$new_pos$new_prev$new_runs";
  };
  require MyFSM;
  my $fsm = MyFSM->new;
  my @pending = ('0000');
  $fsm->set_initial_states(@pending);
  while (@pending) {
    my $state = shift @pending;
    next if $fsm->has_state($state);
    foreach my $bit (0,1) {
      my $to = $cppr_transition->($state,$bit);
      $fsm->add_transition($state, $bit, $to);
      push @pending, $to;
    }
  }
  foreach my $state ($fsm->states) {
    my $to = $fsm->transition($state,0);
    $to = $fsm->transition($to,0);
    $to = $fsm->transition($to,0);
    my ($carry,$pos,$prev,$runs) = split //, $to;
    if ($runs) {
      $fsm->add_accepting($state);
    }
  }
  $fsm->show;
  $fsm->simplify;
  foreach my $state ($fsm->states) {
    $fsm->rename_state($state, "s$state");
  }
  # $fsm->view;
  # $fsm->print_tikz;

  require FLAT;
  my $f = $fsm->as_FLAT;
  # $f = $f->MyFLAT::minimize(verbose=>1);
  $f->MyFLAT::view;
  # $f = $f->MyFLAT::reverse;
  # $f = $f->MyFLAT::minimize(verbose=>1);
  # $f = $f->MyFLAT::rename_breadth_first;
  # $f->MyFLAT::view;

  # my $rev = $fsm->reverse;
  # $rev->simplify;
  # $rev->renumber_states;
  # $rev->view;

  # no state 0010       even pos, negabit 1 below, always carry
  # no state 0011
  # no state 1110        odd pos, negabit 1 below, never carry
  # no state 1111
  foreach my $s (0 .. 15) {
    my $state = sprintf '%04b', $s;
    if (! $fsm->has_state("s$state")) {
      print "no state $state\n";
    }
  }

  my $bad = 0;
  foreach my $n (0 .. 1<<13) {
    my $n2rev = reverse(cnv($n,10,2));
    foreach my $zeros (0 .. 4) {
      my $by_flat = $f->contains($n2rev) ? 1 : 0;
      my $by_func = negabinary_runs_parity($n);
      unless ($by_flat == $by_func) {
        print "n=$n $n2rev $by_flat $by_func\n";
        exit if $bad++ > 10;
      }
      $n2rev .= '0';
    }
  }
  exit 0;
}


#------------------------------------------------------------------------------
# negabinary 1s parity

{
  foreach my $n (0 .. 256) {
    my $n2rev = reverse(cnv($n,10,2));
    foreach my $zeros (0 .. 4) {
      my $by_flat = $negabinary_1s_parity_LtoH_flat->contains($n2rev) ? 1 : 0;
      my $by_func = negabinary_1s_parity($n);
      unless ($by_flat == $by_func) {
        die;
      }
      $n2rev .= '0';
    }
  }
}

{
  my $f = $negabinary_1s_parity_LtoH_flat;
  MyFLAT::FLAT_show_breadth($f, 4, 'lowtohigh');
  {
    # those ending in a 0 strip that 0 gives original
    my $a = $f->intersect($f, FLAT::Regex->new('(0|1)* 0')->as_dfa);
    my @pred = $a->MyFLAT::predecessors([$a->get_accepting],0);
    ### @pred
    $a->unset_accepting($a->get_states);
    $a->set_accepting(@pred);
    MyFLAT::FLAT_check_is_equal($a,$f);
  }
  # $f->MyFLAT::view;

  my $h = $negabinary_1s_parity_LtoH_flat->MyFLAT::reverse;
  $h = $h->MyFLAT::minimize(verbose=>1);
  $h = $h->MyFLAT::rename_breadth_first;
  $h->MyFLAT::reverse->equals($negabinary_1s_parity_LtoH_flat) or die;
  MyFLAT::FLAT_show_breadth($h, 4);

  $h = $h->MyFLAT::FLAT_rename(states_map => { 0=>0,
                                               1=>1,
                                               2=>2,
                                               5=>3,
                                               10=>4,
                                               3=>5,
                                               7=>6,
                                               8=>7,
                                               6=>8,
                                               4=>9,
                                               11=>10,
                                               9=>11 });
  $h->MyFLAT::reverse->equals($negabinary_1s_parity_LtoH_flat) or die;

  MyFLAT::FLAT_print_gp_inline_table($h, 'A269027_HtoL_state_table');

  # $h->MyFLAT::view;
  # MyFLAT::FLAT_print_tikz($h);
  my @accepting = $h->get_accepting;
  print "accepting ", join(',',@accepting), "\n";


  {
    # those starting in a 0 strip that 0 gives original
    my $a = $h->intersect(FLAT::Regex->new('0 (0|1)*')->as_dfa);
    my @succ = $a->successors($a->get_starting,0);
    ### @succ
    $a = $a->MyFLAT::as_nfa;
    $a->unset_starting($a->get_states);
    $a->set_starting(@succ);
    $a = $a->as_dfa;
    MyFLAT::FLAT_check_is_equal($a,$h);
  }

  # $f->set_accepting(@accepting);
  # print "% accepting ", join(',',@accepting), "\n";
  #
  # @accepting = $f->MyFLAT::predecessors(\@accepting, 0);
  # print "% accepting ", join(',',@accepting), "\n";
  # @accepting = $f->MyFLAT::predecessors(\@accepting, 0);
  # print "% accepting ", join(',',@accepting), "\n";
  exit 0;
}
#------------------------------------------------------------------------------
